var request = require('request');
var strftime = require('strftime');
var redis = require('redis');
var winston = require('winston');
var fs = require('fs');
var fsextra = require('fs-extra');
var ini = require('ini');
var redisClient;

var max_retry = 10;
var sleep_before_retry = 3000;

var matrix = {};
var config;
var partner;
var logger = {
    info: function(message) {
        console.log(message);
    },
    warn: function(message) {
        console.log(message);
    },
    err: function(message) {
        console.log(message);
    },
}

var trx_statuses = {};
var available_products = [];
var pending_list = [];

function unaliasResponseCode(response_code, config_responsecode_alias) {
    if (config_responsecode_alias == undefined && config && config.h2h_out && config.h2h_out.responsecode_alias) {
        config_responsecode_alias = config.h2h_out.responsecode_alias;
    }
    
    if (!config_responsecode_alias) {
        return response_code;
    }
    
    items = config_responsecode_alias.split(',');
    items_count = items.length;
    
    for (var i=0; i < items_count; i++) {
        codes = items[i].split(':');
        
        if (codes.length <= 0) { continue; }
        
        if (response_code == codes[0]) {
            logger.info('Change response code from ' + codes[0] + ' to ' + codes[1]);
            return codes[1];
        }
    }
    
    return response_code;
}

function pullCity() {
    var url = config.globals.aaa_host + '/pull_city';
    logger.info('Pull cities from AAA - ' + url);
    request(url, function (error, response, body) {
        if (!error && response.statusCode == 200) {
            //console.log('city=' + body);
        } else {
            logger.warn('Error in pull city');
        }
    });
}

function pullProduct() {
    var url = config.globals.aaa_host + '/pull_product?opr_name=' + config.globals.operators;
    logger.info('Pull products from AAA - ' + url);

    request(url, function (error, response, body) {
        if (error || response.statusCode != 200) {
            logger.warn('Error in pull products');
            return;
        }

        var productsAndOperators = body.split(';');
        var productsCount = productsAndOperators.length;

        for (var i=0; i < productsCount; i++) {
            var product = productsAndOperators[i].split(',', 1)[0];
            available_products.push(product);
        }
        //console.log(available_products);
    });
}

function incrementTrxCount() {
    if (!matrix.counter) {
        matrix.counter = {};
    }
    
    if (!matrix.counter.trx) {
        matrix.counter.trx = {};
    }
    
    if (!matrix.counter.trx.count) {
        matrix.counter.trx.count = 0;
    }
    
    return ++matrix.counter.trx.count;
}

function incrementActiveTrxCount() {
    if (!matrix.counter) {
        matrix.counter = {};
    }
    
    if (!matrix.counter.trx) {
        matrix.counter.trx = {};
    }
    
    if (!matrix.counter.trx.active) {
        matrix.counter.trx.active = 0;
        matrix.counter.trx.active_max = 0;
    }
    
    matrix.counter.trx.active_max = Math.max(matrix.counter.trx.active_max, ++matrix.counter.trx.active);
    return matrix.counter.trx.active;
}

function decrementActiveTrxCount() {
    if (!matrix.counter) {
        matrix.counter = {};
    }
    
    if (!matrix.counter.trx) {
        matrix.counter.trx = {};
    }
    
    if (!matrix.counter.trx.active) {
        matrix.counter.trx.active = 1;
    }
    
    return --matrix.counter.trx.active;
}

function prepareMatrix(_config) {
    var m = {
        config_dirty: 0,
        pause: 0,
        counter: {
            trx: {
                count: 0,
                active: 0,
                active_max: 0,
            }
        }
    };
    
    try {
        if (_config.globals.pause_on_start.toUpperCase() == 'YES') {
            m.pause = 1;
        }
    }
    catch(err) {
        console.log('No entry for config.pause_on_start, run on start automatically');
    }
    
    return m;
}

function pull() {
    var url = config.globals.aaa_host
        + '/pull?city=ALL&nom=' + config.globals.products
        + '&chip_info=' + config.globals.gateway_name;

    //console.log('AAA PULL - ' + url);
    request(url, function (error, response, body) {
        if (!error && response.statusCode == 200) {
            if (body == 'NONE') {
                return;
            }
            
            logger.info('AAA give a task', {body: body});
            
            var result = body.split(';');
            if (result[0] != 'OK') {
                return;
            }
            
            //incrementTrxCount();
            //incrementActiveTrxCount();
            
            var task = {};
            task['requestId'] = result[1];
            task['timestamp'] = result[3];
            task['destination'] = result[4];
            task['member'] = result[5];
            task['product'] = result[7];
            
            pushTaskToPendingList(task);

            if (config.products[task['product']] !== undefined) {
                task['remoteProduct'] = config.products[task['product']];
            } else {
                task['remoteProduct'] = task['product'];
            }
            
            saveTaskToRedis(task);
            publishTaskToRedis(task);
            
            logger.info('New task', {task: task});

            partner.topupRequest(task);

        } else {
            logger.warn('Error in pull task');
            return;
        }
    });
}

function publishTaskToRedis(task) {
    try {
        var key = 'kimochi.aaa_pull.gw:' + config.globals.gateway_name;
        
        redisClient.publish(key + '.json', JSON.stringify(task));
        
        var plain_text = task['timestamp'] + ' - ' + task['member'] + ' isi ' + task['product'] + ' ke ' + task['destination'];
        redisClient.publish(key + '.text', plain_text);
    }
    catch(err) {
        logger.warn('Exception on publish task to redis', {err: err});
    }
}

function getAaaTaskRedisKey(requestId) {
    return config.globals.gateway_name + '.aaa_task.' + requestId;
}

function saveTaskToRedis(task) {
    try {
        var key = getAaaTaskRedisKey(task['requestId']);
        
        redisClient.set(key, JSON.stringify(task));
        redisClient.expire(key, 3600*24*7);
    }
    catch(err) {
        logger.warn('Exception on save task to redis', {err: err});
    }
}

function pushTaskToPendingList(task) {
    var i = pending_list.indexOf(task['requestId']);
    if (i >= 0) {
        return;
    }
    pending_list.push(task['requestId']);
    setTimeout(deleteRequestIdFromPendingList, 1000 * 3600 * 24, task['requestId']);
}

function getStockFromMessage(message, stock_regex, stock_regex_match_index) {
    if (!stock_regex) {
        return null;
    }
    
    if (!stock_regex_match_index) {
        return null;
    }
    
    
    var matches;
    try {
        var re = new RegExp(stock_regex);
        matches = message.match(re);
        var result = matches[stock_regex_match_index];
        
        return Number(result);
    }
    catch(err) {
        logger.warn('No stock information available', {catch_err: err, re: stock_regex, re_match_index: stock_regex_match_index, message: message, matches: matches});
        return null;
    }
}

function deleteRequestIdFromPendingList(request_id) {
    var i = pending_list.indexOf(request_id);
    if (i < 0) {
        return;
    }
    pending_list.splice(i, 1);
}

function getPendingList() {
    return pending_list;
}

function isPaused() {
    return matrix.pause;
}

function pause() {
    matrix.pause = 1;
}

function resume() {
    matrix.pause = 0;
}

function pullLoop() {
    //if (!config.globals.pause && config.globals.products) {
    if (!isPaused() && config.globals.products) {
        pull();
    }
    
    setTimeout(pullLoop, config.globals.interval);
}

function deleteTrxStatus(requestId) {
    delete trx_statuses[requestId];
}

function isEnoughBalance(balance) {
    var min_balance = config.globals.min_balance;
    
    min_balance = Number(min_balance);
    if (!min_balance) {
        return true;
    }
    
    balance = Number(balance);
    if (!balance) {
        return true;
    }
    
    return balance >= min_balance;
}

function detectDuplicateTrxFromMessage(message) {
    if (!config.globals.duplicate_regex) {
        return false;
    }
    
    var re = new RegExp(config.globals.duplicate_regex);
    if (message.search(re) >= 0) {
        return true;
    }
}

function getSnFromMessage(message) {
    var sn_match = message.match(/^SN=.*?;/);
    if (!sn_match) {
        return null;
    }
    
    try {
        var sn = sn_match[0];
        sn = sn.replace(/^SN=/, '');
        sn = sn.replace(/;$/, '');
        
        if (!sn) return null;
        
        return sn;
    }
    catch(err) {
        return null;
    }
}

function removeGarbageFromSnMessage(message, regex_pattern) {
    if (!regex_pattern) {
        if (!config.globals.garbages_on_sn_regex) {
            return message;
        }
        regex_pattern = config.globals.garbages_on_sn_regex;
    }
    
    var original_sn = getSnFromMessage(message);
    if (!original_sn) {
        return message;
    }
    
    var re = new RegExp(regex_pattern, 'g');
    new_sn = original_sn.replace(re, '');
    
    return message.replace(original_sn, new_sn);
}

function callbackReport(requestId, responseCode, message, retry) {
    // detect duplicate trx from message
    if (responseCode == '00' && detectDuplicateTrxFromMessage(message)) {

        callbackReport(requestId, '55', message);
        return;

    } else {

        var balance = getBalanceFromMessage(message);
        if (balance) {
            updateBalance(balance);
        }
    }
    
    try {
        message = message.replace(/(\d)=(\d)/g, "$1 = $2");
    } 
    catch(err) {
        
    }
    
    // ** obsoleted, diganti removeGarbageFromSnMessage)
    // cleanup some dirty SN (sometimes R21 return ":" as first char SN)
    // message = message.replace(/^SN=:/, "SN=");
    
    if ((responseCode == '68') && isFinishedTrx(requestId)) {
        logger.info('Ignoring finished trx (%d)', requestId);
        return;
    } else {
        trx_statuses[requestId] = responseCode;
        setTimeout(deleteTrxStatus, 24 * 3600 * 1000, requestId);
    }
    
    if (retry === undefined) {
        retry = max_retry;
    }
    
    responseCode = unaliasResponseCode(responseCode);
    if (responseCode == '00') {
        message = removeGarbageFromSnMessage(message);
    }
    
    var timestamp = strftime('%Y%m%d%H%M%S');
    var options = {
        url: config.globals.aaa_host + '/topup',
        qs: {
            trans_id: requestId,
            trans_date: timestamp,
            resp_code: responseCode,
            ussd_msg: config.globals.gateway_name + '$' + message
        }
    };
    
    try {
        var redis_key = 'echi.topup_report.gw:' + config.globals.gateway_name;
        redisClient.publish(redis_key + '.json', JSON.stringify(options.qs));
        
        var redis_message = timestamp + ' - ' + responseCode + ' - ' + message;
        redisClient.publish(redis_key + '.message', redis_message);
    }
        catch(err) {
    }

    logger.info('Report to AAA', {qs: options.qs});
    request(options, function (error, response, body) {
        if (error || response.statusCode != 200) {
            logger.warn('Error report to AAA');
            
            if (retry) {
                logger.info('Retrying to report to AAA (%d)', retry);
                callbackReport(requestId, responseCode, message, retry - 1);
            }
        }
    });
    
    if (responseCode != '68') {
        deleteRequestIdFromPendingList(requestId);
    }
    
    if (responseCode == '00') {
        updateStockFromMessage(requestId, message);
    }
}

function updateStockFromMessage(requestId, message) {
    if (!config.globals.stock_regex || !config.globals.stock_regex_index) {
        return;
    }
    
    var stock = getStockFromMessage(message, config.globals.stock_regex, config.globals.stock_regex_index);
    
    if (stock == null) {
        return;
    }
    
    var task_key = getAaaTaskRedisKey(requestId);
    redisClient.get(task_key, function(err, _task) {
        if (err) {
            logger.warn('Error on getting task from redis', {err: err});
            return;
        }
        
        var task;
        try {
            task = JSON.parse(_task);
        }
        catch(json_err) {
            logger.warn('Error parsing json on updateStockFromMessage', {err: json_err});
        }
        
        if (!task) {
            return;
        }
        
        //var stock_key = config.globals.gateway_name + '.product:' + task['product'] + '.stock';
        try {
            var stock_key = 'balance.gw:' + config.globals.gateway_name + '.' + task['remoteProduct'].toUpperCase();
            logger.info('Saving stock information: ' + stock);
            redisClient.set(stock_key, stock);
            redisClient.expire(stock_key, 30 * 24 * 3600 * 10);
        }
        catch(set_err) {
            logger.warn('Exception on saving stock to redis', {err: set_err});
        }
        
    });
}

function createRedisClient() {
    try {
        redisClient = redis.createClient(config.globals.redis_port, config.globals.redis_host);
    } catch(err) {
        logger.info("Error creating redis client");
    }
}

function start(_config, _partner, options) {
    config = _config;
    partner = _partner;
    
    if (options && options.matrix) {
        matrix = options.matrix;
    } else {
        matrix = {
            pause: 0
        };
    }
    
    if (options && options.logger) {
        logger = options.logger;
    } else {
        logger = new winston.Logger({
            transports: [
              new (winston.transports.Console)()
            ]
        });
    }

    createRedisClient();
    
    setTimeout(pullLoop, 10 * 1000);
}

function getBalanceFromMessage(message, balance_regex) {
    if (!balance_regex) {
        if (config && config.globals && config.globals.balance_regex) {
            balance_regex = config.globals.balance_regex;
        }
    }
    
    if (!balance_regex) {
        return;
    }
    
    try {
        var re = new RegExp(balance_regex);
        var matches = message.match(re);

        var result = matches[1];
        result = result.replace(/\./g, '');
        result = result.replace(/,/g, '');
        
        
        return Number(result);
    }
    catch(err) {
        return;
    }
}

function updateBalance(balance) {
    
    if (!isEnoughBalance(balance)) {
        logger.info("Suspend on low balance. PAUSED");
        pause();
    }
    
    try {
        var key = 'balance.gw:' + config.globals.gateway_name;
        var value = Number(balance);
        if (value) {
            redisClient.set(key, value);
            redisClient.expire(key, 3600*24*7);
        }
    }
    catch(err) {
        logger.warn('Update balance failed: ' + err);
    }
}

function isFinishedTrx(req_id) {
    if (trx_statuses[req_id] === undefined) {
        return false;
    }
    
    if (trx_statuses[req_id] == '68') {
        return false;
    }
    
    return true;
}

function getProductCode(product) {
    var result = product.match(/^\D+/);
    
    try {
        return result[0];
    }
    catch(err) {
        return;
    }
}

function getProductNominal(product) {
    var result = product.match(/\d+/);
    
    try {
        return Number(result[0]);
    } 
    catch(err) {
        return;
    }
}

function sortProductsArray(products) {
    products.sort(function(a, b) {
        var a_code = getProductCode(a);
        var b_code = getProductCode(b);
        
        var compare_by_code = a_code.localeCompare(b_code);
        if (compare_by_code) {
            return compare_by_code;
        }
        
        var a_nominal = getProductNominal(a);
        var b_nominal = getProductNominal(b);
        return a_nominal - b_nominal;
    });
    
    return products;
}

function addProduct(product, products) {
    logger.info("Add product " + product, {product: product, products: products});
    if (products.indexOf(product) < 0) {
        products.push(product);
        
        if (products.length > 1) {
            products = sortProductsArray(products);
        }
        matrix.config_dirty = 1;
    }
    return products;
}

function delProduct(product, products) {
    logger.info("Delete product " + product, {product: product, products: products});
    var pos = products.indexOf(product);
    
    if (pos >= 0) {
        products.splice(pos, 1);
        
        if (products.length > 1) {
            products = sortProductsArray(products);
        }
        matrix.config_dirty = 1;
    }
    return products;
}

function backupAndSaveConfig() {
    fsextra.copy('config.ini', 'config.ini.backup_' + strftime('%F_%H%M%S', new Date()), function(err) {
        fs.writeFileSync('./config.ini', ini.stringify(config));
        matrix.config_dirty = 0;
    });
}

function loadConfig() {
    console.log('Old config: ');
    console.log(config);
    config = {};
    config = ini.parse(fs.readFileSync('config.ini', 'utf-8'));
    matrix.config_dirty = 0;
    console.log('New config: ');
    console.log(config);
}

exports.start = start;
exports.callbackReport = callbackReport;
exports.unaliasResponseCode = unaliasResponseCode;
exports.updateBalance = updateBalance;
exports.getBalanceFromMessage = getBalanceFromMessage;
exports.getPendingList = getPendingList;
exports.getProductCode = getProductCode;
exports.getProductNominal = getProductNominal;
exports.sortProductsArray = sortProductsArray;
exports.addProduct = addProduct;
exports.delProduct = delProduct;
exports.pause = pause;
exports.resume = resume;
exports.prepareMatrix = prepareMatrix;
exports.incrementTrxCount = incrementTrxCount;
exports.incrementActiveTrxCount = incrementActiveTrxCount;
exports.decrementActiveTrxCount = decrementActiveTrxCount;
exports.backupAndSaveConfig = backupAndSaveConfig;
exports.loadConfig = loadConfig;
exports.getStockFromMessage = getStockFromMessage;
exports.getSnFromMessage = getSnFromMessage;
exports.removeGarbageFromSnMessage = removeGarbageFromSnMessage;
