var assert = require("assert");


describe('aaa', function() {
    var aaa = require('./aaa');
    
    describe("#unaliasResponseCode()", function() {
        it('should return 68', function() {
            assert.equal('68', aaa.unaliasResponseCode('01', '01:68'));
        });
        
        it('should return 68', function() {
            assert.equal('68', aaa.unaliasResponseCode('68', '01:68'));
        });
        
        it('should return 00', function() {
            assert.equal('00', aaa.unaliasResponseCode('00', '01:68'));
        });
        
        it('should return 40', function() {
            assert.equal('40', aaa.unaliasResponseCode('40', ''));
        });
        
        it('should return 40', function() {
            assert.equal('40', aaa.unaliasResponseCode('40', ''));
        });
        
        it('should return 40', function() {
            assert.equal('40', aaa.unaliasResponseCode('40'));
        });
    
    });
    
    describe("#getBalanceFromMessage()", function() {
        it ('should return 4782801', function() {
            var message = 'Sukses! , Topup S20H Ke 082136047702, TrxID: 17699291 SN: 0031000909857657 Saldo Rp. 4,782,801=4,802,701-19,900.INFO: TSEL 10 CLOSE.NUSAPULSA';
            assert.equal(aaa.getBalanceFromMessage(message, 'Saldo Rp\. (.+?)='), 4782801);
            assert.equal(aaa.getBalanceFromMessage('2016-02-15 17:00 TRX HSM25 088212551565 BERHASIL S/N: 392745165428  Harga:24575 Saldo: 2442265', 'Saldo: (\\d+?)$'), 2442265);
        });
    });
    
    describe("#getProductCode()", function() {
        it ('should return SP', function() {
            assert.equal('SP', aaa.getProductCode('SP'));
            assert.equal('SP', aaa.getProductCode('SP20'));
            assert.equal('SP', aaa.getProductCode('SP30'));
            assert.equal('SP', aaa.getProductCode('SP10F'));
            assert.equal(undefined, aaa.getProductCode(''));
            assert.equal(undefined, aaa.getProductCode('23'));
        });
    });
    
    describe("#getProductNominal()", function() {
        it ('should return 20', function() {
            assert.equal(20, aaa.getProductNominal('SP20'));
            assert.equal(20, aaa.getProductNominal('SP20F'));
            assert.equal(undefined, aaa.getProductNominal('SP'));
        });
    });
    
    describe("#setProductsArray()", function() {
        it ('should ok', function() {
            assert.deepEqual(aaa.sortProductsArray(["TR100", "TR10","SP20","SP100", "SP10"]), ["SP10","SP20","SP100", "TR10", "TR100"]);
            assert.deepEqual(aaa.sortProductsArray([]), []);
        });
    });
    
    describe('#getStockFromMessage()', function() {
        it ('should return correct stock', function() {
            assert.equal(
                105,
                aaa.getStockFromMessage(
                    'SN=0041000371763354; Berhasil 14/25/1103 : S20H ke 085215921744 SN : 41000371763354 Reff:4109005271537231, Saldo Rp. 10.984.513 (105) = Rp. 11.004.513 - Rp. 20.000.SA_POIN',
                    '\\d \\((\\d+)\\) =',
                    1
                )
            );
        });
    });
    
    describe('#getSnFromMessage()', function() {
        it ('should return null if message does not contain sn', function() {
            assert.equal(aaa.getSnFromMessage('XXXX, XR10, KE 8888888 BERHASIL , SN=044490129220452;, HARGA=10,275 , SISA SALDO RP. 1,264,350 REBET ANDA RP.0 - NEW XL,AXIS DENOM 15, KODE X15'), null);
            assert.equal(aaa.getSnFromMessage('SN=;XXXX, XR10, KE 8888888 BERHASIL , SN=044490129220452;, HARGA=10,275 , SISA SALDO RP. 1,264,350 REBET ANDA RP.0 - NEW XL,AXIS DENOM 15, KODE X15'), null);
        });
        
        it ('should return valid sn', function() {
            assert.equal(aaa.getSnFromMessage('SN=044490129220452; XXXX, XR10, KE 8888 BERHASIL , SN=044490129220452;, HARGA=10,275 , SISA SALDO RP. 1,264,350 REBET ANDA RP.0 - NEW XL,AXIS DENOM 15, KODE X15'), 044490129220452);
        });
    });
    
    describe('#removeGarbageFromSnMessage()', function() {
        it ('should return original message', function() {
            assert.equal(aaa.removeGarbageFromSnMessage('XXXX, XR10, KE 8888888 BERHASIL , SN=044490129220452;, HARGA=10,275 , SISA SALDO RP. 1,264,350 REBET ANDA RP.0 - NEW XL,AXIS DENOM 15, KODE X15', '[,:;]'), 'XXXX, XR10, KE 8888888 BERHASIL , SN=044490129220452;, HARGA=10,275 , SISA SALDO RP. 1,264,350 REBET ANDA RP.0 - NEW XL,AXIS DENOM 15, KODE X15');
            assert.equal(aaa.removeGarbageFromSnMessage('SN=;XXXX, XR10, KE 8888888 BERHASIL , SN=044490129220452;, HARGA=10,275 , SISA SALDO RP. 1,264,350 REBET ANDA RP.0 - NEW XL,AXIS DENOM 15, KODE X15', '[,:;]'), 'SN=;XXXX, XR10, KE 8888888 BERHASIL , SN=044490129220452;, HARGA=10,275 , SISA SALDO RP. 1,264,350 REBET ANDA RP.0 - NEW XL,AXIS DENOM 15, KODE X15');
            assert.equal(aaa.removeGarbageFromSnMessage('SN=044490129220452; XXXX, XR10, KE 8888 BERHASIL , SN=044490129220452;, HARGA=10,275 , SISA SALDO RP. 1,264,350 REBET ANDA RP.0 - NEW XL,AXIS DENOM 15, KODE X15', '[,:;]'), 'SN=044490129220452; XXXX, XR10, KE 8888 BERHASIL , SN=044490129220452;, HARGA=10,275 , SISA SALDO RP. 1,264,350 REBET ANDA RP.0 - NEW XL,AXIS DENOM 15, KODE X15');
        });
        
        it ('should return message with clean sn', function() {
            assert.equal(
                aaa.removeGarbageFromSnMessage(
                    'SN=:044490129220452; XXXX, XR10, KE 8888 BERHASIL , SN=044490129220452;, HARGA=10,275 , SISA SALDO RP. 1,264,350 REBET ANDA RP.0 - NEW XL,AXIS DENOM 15, KODE X15', 
                    '[,:;]'
                ),
                'SN=044490129220452; XXXX, XR10, KE 8888 BERHASIL , SN=044490129220452;, HARGA=10,275 , SISA SALDO RP. 1,264,350 REBET ANDA RP.0 - NEW XL,AXIS DENOM 15, KODE X15'
            );
        });
    });
});
